package ru.dbotthepony.kstarbound.client.gl.vertex

import org.lwjgl.opengl.GL46
import org.lwjgl.opengl.GL46.GL_UNSIGNED_INT
import org.lwjgl.opengl.GL46.GL_UNSIGNED_SHORT
import org.lwjgl.opengl.GL46.GL_UNSIGNED_BYTE
import ru.dbotthepony.kstarbound.client.gl.GLType
import ru.dbotthepony.kstarbound.client.gl.VertexBufferObject
import ru.dbotthepony.kstarbound.client.gl.checkForGLError
import ru.dbotthepony.kstarbound.util.writeLEFloat
import ru.dbotthepony.kstarbound.util.writeLEInt
import ru.dbotthepony.kstarbound.util.writeLEShort
import ru.dbotthepony.kvector.util2d.AABB
import java.io.OutputStream
import kotlin.math.cos
import kotlin.math.sin

private fun put(type: Int, memory: OutputStream, value: Int) {
	when (type) {
		GL_UNSIGNED_SHORT -> memory.writeLEShort(value)
		GL_UNSIGNED_BYTE -> memory.write(value)
		else -> memory.writeLEInt(value)
	}
}

/**
 * Класс для построения геометрии для загрузки в память видеокарты.
 */
@Suppress("unchecked_cast")
abstract class AbstractVertexBuilder<out T : AbstractVertexBuilder<T>>(
	val attributes: GLAttributeList,
	val type: GeometryType,
) {
	protected abstract val vertexMemory: OutputStream
	protected abstract val elementMemory: OutputStream

	/**
	 * [GL_UNSIGNED_BYTE], [GL_UNSIGNED_SHORT] или [GL_UNSIGNED_INT]
	 *
	 * Обязана быть статичным числом.
	 *
	 * Если нет, и его надо изменить, то это можно только делать внутри [ensureIndexCapacity].
	 *
	 * Подкласс обязан самостоятельно изменить тип существующих элементов внутри [elementMemory]
	 */
	abstract val elementIndexType: Int

	protected abstract fun ensureIndexCapacity()

	protected abstract fun resetMemory()

	fun begin() {
		inVertex = false
		attributeIndex = 0
		elementIndexOffset = 0
		vertexCount = 0
		indexCount = 0
		resetMemory()
	}

	private var inVertex = false
	private var attributeIndex = 0
	protected var elementIndexOffset = 0
		private set

	var vertexCount = 0
		private set

	var indexCount = 0
		private set

	protected fun checkBounds() {
		if (attributeIndex >= attributes.size) {
			throw IndexOutOfBoundsException("Tried to add new attribute when already added all attributes")
		}
	}

	fun expect(type: GLType) {
		checkBounds()

		if (attributes[attributeIndex].glType != type) {
			throw IllegalStateException("Expected attribute type $type, got ${attributes[attributeIndex].name}[${attributes[attributeIndex].glType}] (at position $attributeIndex)")
		}
	}

	fun expect(name: String) {
		checkBounds()

		if (attributes[attributeIndex].name != name) {
			throw IllegalStateException("Expected attribute name $name, got ${attributes[attributeIndex].name}[${attributes[attributeIndex].glType}] (at position $attributeIndex)")
		}
	}

	protected abstract fun doUpload(vbo: VertexBufferObject, ebo: VertexBufferObject, drawType: Int)

	fun upload(vbo: VertexBufferObject, ebo: VertexBufferObject, drawType: Int = GL46.GL_DYNAMIC_DRAW) {
		require(vbo.isArray) { "$vbo is not an array" }
		require(ebo.isElementArray) { "$vbo is not an element array" }

		end()

		check(elementVertices == 0) { "Not fully built vertex element ($type requires ${type.elements} vertex points to be present, yet last strip has only $elementVertices elements)" }

		doUpload(vbo, ebo, drawType)
		checkForGLError()
	}

	private var elementVertices = 0

	fun end() {
		if (inVertex) {
			inVertex = false

			if (attributeIndex != attributes.size) {
				throw IllegalStateException("Unfinished vertex, we are at $attributeIndex, while we have ${attributes.size} attributes")
			}

			vertexCount++
			elementVertices++

			if (elementVertices == type.elements) {
				ensureIndexCapacity()

				elementVertices = 0
				val elementMemory = elementMemory
				val elementIndexType = elementIndexType

				for (index in type.indices) {
					put(elementIndexType, elementMemory, index + elementIndexOffset)
				}

				elementIndexOffset += type.elements
				indexCount += type.indices.size
			}
		}
	}

	fun vertex(): T {
		end()
		inVertex = true
		attributeIndex = 0
		return this as T
	}

	fun pushVec4f(x: Float, y: Float, z: Float, w: Float): T {
		expect(GLType.VEC4F)
		val memory = vertexMemory
		memory.writeLEFloat(x)
		memory.writeLEFloat(y)
		memory.writeLEFloat(z)
		memory.writeLEFloat(w)
		attributeIndex++
		return this as T
	}

	fun pushVec3f(x: Float, y: Float, z: Float): T {
		expect(GLType.VEC3F)
		val memory = vertexMemory
		memory.writeLEFloat(x)
		memory.writeLEFloat(y)
		memory.writeLEFloat(z)
		attributeIndex++
		return this as T
	}

	fun pushVec2f(x: Float, y: Float): T {
		expect(GLType.VEC2F)
		val memory = vertexMemory
		memory.writeLEFloat(x)
		memory.writeLEFloat(y)
		attributeIndex++
		return this as T
	}

	fun push(value: Float): T {
		expect(GLType.FLOAT)
		vertexMemory.writeLEFloat(value)
		attributeIndex++
		return this as T
	}
}

fun <T : AbstractVertexBuilder<T>> T.shadowLine(x0: Float, y0: Float, x1: Float, y1: Float): T {
	vertex()
	pushVec4f(x0, y0, x1, y1)
	pushVec2f(0f, 0f)

	vertex()
	pushVec4f(x0, y0, x1, y1)
	pushVec2f(0f, 1f)

	vertex()
	pushVec4f(x0, y0, x1, y1)
	pushVec2f(1f, 1f)

	vertex()
	pushVec4f(x0, y0, x1, y1)
	pushVec2f(1f, 0f)

	return this
}

fun <T : AbstractVertexBuilder<T>> T.dfShadowLine(x0: Float, y0: Float, x1: Float, y1: Float): T {
	shadowLine(x0, y0, x1, y1)
	shadowLine(x1, y1, x0, y0)
	return this
}

fun <T : AbstractVertexBuilder<T>> T.shadowQuad(x0: Float, y0: Float, x1: Float, y1: Float): T {
	shadowLine(x0, y0, x1, y0)
	shadowLine(x1, y0, x1, y1)
	shadowLine(x1, y1, x0, y1)
	shadowLine(x0, y1, x0, y0)
	return this
}

// Помощники
fun <T : AbstractVertexBuilder<T>> T.quad(
	x0: Float,
	y0: Float,
	x1: Float,
	y1: Float,
	lambda: QuadVertexTransformer = EMPTY_VERTEX_TRANSFORM
): T {
	check(type.elements == 4) { "Currently building $type" }

	lambda(vertex().pushVec2f(x0, y0), 0).end()
	lambda(vertex().pushVec2f(x1, y0), 1).end()
	lambda(vertex().pushVec2f(x0, y1), 2).end()
	lambda(vertex().pushVec2f(x1, y1), 3).end()

	return this
}

fun <T : AbstractVertexBuilder<T>> T.quadRotated(
	x0: Float,
	y0: Float,
	x1: Float,
	y1: Float,
	x: Float,
	y: Float,
	angle: Double,
	lambda: QuadVertexTransformer = EMPTY_VERTEX_TRANSFORM
): T {
	check(type.elements == 4) { "Currently building $type" }

	val s = sin(angle).toFloat()
	val c = cos(angle).toFloat()

	lambda(vertex().pushVec2f(x + x0 * c - s * y0, y + s * x0 + c * y0), 0).end()
	lambda(vertex().pushVec2f(x + x1 * c - s * y0, y + s * x1 + c * y0), 1).end()
	lambda(vertex().pushVec2f(x + x0 * c - s * y1, y + s * x0 + c * y1), 2).end()
	lambda(vertex().pushVec2f(x + x1 * c - s * y1, y + s * x1 + c * y1), 3).end()

	return this
}

fun <T : AbstractVertexBuilder<T>> T.quad(aabb: AABB, lambda: QuadVertexTransformer = EMPTY_VERTEX_TRANSFORM): T {
	return quad(
		aabb.mins.x.toFloat(),
		aabb.mins.y.toFloat(),
		aabb.maxs.x.toFloat(),
		aabb.maxs.y.toFloat(),
		lambda
	)
}

fun <T : AbstractVertexBuilder<T>> T.quadZ(
	x0: Float,
	y0: Float,
	x1: Float,
	y1: Float,
	z: Float,
	lambda: QuadVertexTransformer = EMPTY_VERTEX_TRANSFORM
): T {
	check(type.elements == 4) { "Currently building $type" }

	lambda(vertex().pushVec3f(x0, y0, z), 0).end()
	lambda(vertex().pushVec3f(x1, y0, z), 1).end()
	lambda(vertex().pushVec3f(x0, y1, z), 2).end()
	lambda(vertex().pushVec3f(x1, y1, z), 3).end()

	return this
}

fun <T : AbstractVertexBuilder<T>> T.quadRotatedZ(
	x0: Float,
	y0: Float,
	x1: Float,
	y1: Float,
	z: Float,
	x: Float,
	y: Float,
	angle: Double,
	lambda: QuadVertexTransformer = EMPTY_VERTEX_TRANSFORM
): T {
	check(type.elements == 4) { "Currently building $type" }

	val s = sin(angle).toFloat()
	val c = cos(angle).toFloat()

	lambda(vertex().pushVec3f(x + x0 * c - s * y0, y + s * x0 + c * y0, z), 0).end()
	lambda(vertex().pushVec3f(x + x1 * c - s * y0, y + s * x1 + c * y0, z), 1).end()
	lambda(vertex().pushVec3f(x + x0 * c - s * y1, y + s * x0 + c * y1, z), 2).end()
	lambda(vertex().pushVec3f(x + x1 * c - s * y1, y + s * x1 + c * y1, z), 3).end()

	return this
}
